"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PhantomDependencyPlugin = void 0;
const tslib_1 = require("tslib");
const utils_1 = require("@morjs/utils");
const path_1 = tslib_1.__importDefault(require("path"));
const constants_1 = require("../constants");
/**
 * 检测项目中的幽灵依赖
 * 以下情况直接跳过该文件的检测
 * - 生产环境
 * - phantomDependency 为 false 时
 * - 该文件为 node_modules 文件
 * - 文件不在当前 pwd 运行路径内
 */
class PhantomDependencyPlugin {
    constructor() {
        this.name = 'PhantomDependencyPlugin';
    }
    apply(runner) {
        runner.hooks.webpackWrapper.tap(this.name, (webpackWrapper) => {
            this.webpackWrapper = webpackWrapper;
        });
        const usedDependencies = {};
        runner.hooks.beforeRun.tap(this.name, () => {
            const { phantomDependency } = runner.userConfig || {};
            if (!phantomDependency)
                return;
            runner.hooks.scriptParser.tap(this.name, (transformers, options) => {
                const fileInfoPath = options.fileInfo.path || '';
                if (!constants_1.NODE_MODULE_REGEXP.test(fileInfoPath) &&
                    fileInfoPath.includes(runner.getCwd())) {
                    this.handlePhantomTransformer(transformers, fileInfoPath, usedDependencies);
                }
                return transformers;
            });
            runner.hooks.sjsParser.tap(this.name, (transformers, options) => {
                const fileInfoPath = options.fileInfo.path || '';
                if (!constants_1.NODE_MODULE_REGEXP.test(fileInfoPath) &&
                    fileInfoPath.includes(runner.getCwd())) {
                    this.handlePhantomTransformer(transformers, fileInfoPath, usedDependencies);
                }
                return transformers;
            });
        });
        runner.hooks.done.tap(this.name, () => {
            var _a, _b, _c;
            const { alias = {}, srcPaths = [], phantomDependency, externals = [], consumes = [], watch } = (runner.userConfig || {});
            if (!phantomDependency)
                return;
            let allDependencies = { ...this.getPkgDepend(runner.getCwd()) };
            srcPaths.map((srcPath) => {
                allDependencies = { ...allDependencies, ...this.getPkgDepend(srcPath) };
            });
            // 跳过 alias 及以 alias 中以 key 开头的依赖包
            const aliasAll = {
                ...alias,
                ...(_c = (_b = (_a = this.webpackWrapper) === null || _a === void 0 ? void 0 : _a.config) === null || _b === void 0 ? void 0 : _b.resolve) === null || _c === void 0 ? void 0 : _c.alias
            };
            for (const aliasKey in aliasAll) {
                if (aliasKey.endsWith('$')) {
                    aliasAll[aliasKey.substring(0, aliasKey.length - 1)] =
                        aliasAll[aliasKey];
                    delete aliasAll[aliasKey];
                }
            }
            // 跳过在 externals 或 consumes 中配置的包
            const otherDepends = [...(0, utils_1.asArray)(externals), ...consumes].reduce((res, item) => {
                return res.concat(this.getExternalsPkgName(item));
            }, []);
            const table = {
                head: ['依赖名', '引用地址'],
                rows: []
            };
            // 跳过已在 package.json 和 phantomDependency.exclude 中配置的依赖
            for (const depKey in usedDependencies) {
                if (!(phantomDependency['exclude'] || []).includes(depKey) &&
                    !allDependencies[depKey] &&
                    !aliasAll[depKey] &&
                    !aliasAll[depKey.split('/')[0]] &&
                    !otherDepends.includes(depKey)) {
                    table.rows.push([depKey, usedDependencies[depKey]]);
                }
            }
            if (table.rows.length > 0) {
                if (phantomDependency['mode'] === 'error') {
                    utils_1.logger.error('检测到幽灵依赖，请添加到 package.json');
                    utils_1.logger.table(table, 'error');
                    if (!watch) {
                        // 非 watch 状态下，确保 error 模式，可以异常退出
                        const error = new Error();
                        error['isErrorLogged'] = true;
                        error.stack = '';
                        throw error;
                    }
                }
                else {
                    utils_1.logger.warnOnce('检测到幽灵依赖，请添加到 package.json');
                    utils_1.logger.table(table, 'warn');
                }
            }
        });
    }
    // 检查 script 和 sjs 文件 获取 import 和 require 的依赖进行检测
    handlePhantomTransformer(transformers, fileInfoPath, usedDependencies) {
        transformers.before.push((0, utils_1.tsTransformerFactory)((node) => {
            var _a, _b, _c, _d;
            if (!node)
                return node;
            // 处理 import 节点
            if (utils_1.typescript.isImportDeclaration(node) &&
                utils_1.typescript.isStringLiteral(node.moduleSpecifier)) {
                const pkgName = this.getPkgName((_a = node.moduleSpecifier) === null || _a === void 0 ? void 0 : _a.text);
                // 跳过以 . / @/ @@/ 开头的本地文件引用
                if (!/^(\.|\/|@\/|@@\/)/.test(pkgName))
                    usedDependencies[pkgName] = fileInfoPath;
            }
            // 处理 require 节点
            if (utils_1.typescript.isCallExpression(node) &&
                utils_1.typescript.isIdentifier(node.expression) &&
                ((_b = node.expression) === null || _b === void 0 ? void 0 : _b.escapedText) === 'require' &&
                utils_1.typescript.isStringLiteral((_c = node.arguments) === null || _c === void 0 ? void 0 : _c[0])) {
                const pkgName = this.getPkgName((_d = node.arguments) === null || _d === void 0 ? void 0 : _d[0].text);
                // 跳过以 . / @/ @@/ 开头的本地文件引用
                if (!/^(\.|\/|@\/|@@\/)/.test(pkgName))
                    usedDependencies[pkgName] = fileInfoPath;
            }
            return node;
        }));
    }
    /**
     * 获取依赖的 npm 包名
     * @param source - string 被使用依赖
     */
    getPkgName(source) {
        const arr = source.split('/');
        if (source.startsWith('@'))
            return arr[1] ? arr[0] + '/' + arr[1] : arr[0];
        return arr[0];
    }
    /**
     * 获取指定路径 package.json 内的所有依赖
     * @param path 路径
     */
    getPkgDepend(filePath) {
        if (utils_1.fsExtra.existsSync(path_1.default.join(filePath, 'package.json'))) {
            try {
                const pkgJson = utils_1.fsExtra.readJSONSync(path_1.default.join(filePath, 'package.json'));
                return { ...pkgJson.devDependencies, ...pkgJson.dependencies };
            }
            catch (err) {
                utils_1.logger.warn(`${path_1.default.join(filePath, 'package.json')} 文件读取错误: ${err}`);
            }
        }
        return {};
    }
    /**
     * 获取 externals 或 consumes 的依赖名
     * NOTE: 有限支持，externals 可能为函数，这种情况下暂时忽略兼容性
     * @param item 配置子项
     */
    getExternalsPkgName(item) {
        if (typeof item === 'string')
            return item;
        if (utils_1.lodash.isPlainObject(item)) {
            return Object.keys(item);
        }
        return [];
    }
}
exports.PhantomDependencyPlugin = PhantomDependencyPlugin;
//# sourceMappingURL=phantomDependencyPlugin.js.map