"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.preprocess = void 0;
const tslib_1 = require("tslib");
const utils_1 = require("@morjs/utils");
const os_1 = require("os");
const xregexp_1 = tslib_1.__importDefault(require("xregexp"));
const constants_1 = require("../constants");
const JsLikeFileExts = [
    ...constants_1.AllStyleFileTypes,
    ...constants_1.AllConfigFileTypes,
    ...constants_1.AllScriptFileTypes,
    ...constants_1.AllSjsFileTypes,
    '.jsx',
    '.tsx'
];
const XmlLikeFileExts = constants_1.AllTemplateFileTypes;
// 条件编译的正则
const RegexRules = {
    xml: {
        if: {
            start: '[ \t]*<!--[ \t]*#(ifndef|ifdef|if)[ \t]+(.*?)[ \t]*(?:-->|!>)(?:[ \t]*\n+)?',
            end: '[ \t]*<!(?:--)?[ \t]*#endif[ \t]*(?:-->|!>)(?:[ \t]*\n)?'
        }
    },
    js: {
        if: {
            start: '[ \t]*(?://|/\\*)[ \t]*#(ifndef|ifdef|if)[ \t]+([^\n*]*)(?:\\*(?:\\*|/))?(?:[ \t]*\n+)?',
            end: '[ \t]*(?://|/\\*)[ \t]*#endif[ \t]*(?:\\*(?:\\*|/))?(?:[ \t]*\n)?'
        }
    }
};
function normalizeEol(source, indent) {
    if (indent) {
        source = source.replace(/(?:\r?\n)|\r/g, '\n' + indent);
    }
    else {
        source = source.replace(/(?:\r\n)|\r/g, '\n');
    }
    return source;
}
function restoreEol(normalizedSource, originalEol) {
    if (originalEol !== '\n') {
        normalizedSource = normalizedSource.replace(/\n/g, originalEol);
    }
    return normalizedSource;
}
// eslint-disable-next-line @typescript-eslint/ban-types
function getTestTemplate(test) {
    test = test || 'true';
    test = test.trim();
    // force single equals replacement
    test = test.replace(/([^=!])=([^=])/g, '$1==$2');
    test = test.replace(/-/g, '_');
    /* eslint-disable no-new-func */
    return new Function('context', `
    try {
      with (context || {}) {
        return (${test})
      }
    } catch (err) {
      return err
    }
  `);
}
function getDeepPropFromObj(obj, propPath) {
    propPath.replace(/\[([^\]+?])\]/g, '.$1');
    const propPaths = propPath.split('.');
    // fast path, no need to loop if structurePath contains only a single segment
    if (propPaths.length === 1) {
        return obj[propPaths[0]];
    }
    // loop only as long as possible (no exceptions for null/undefined property access)
    propPaths.some(function (pathSegment) {
        obj = obj[pathSegment];
        return obj == null;
    });
    return obj;
}
function testPasses(test, context) {
    const testFn = getTestTemplate(test);
    try {
        return testFn(context, getDeepPropFromObj);
    }
    catch (e) {
        utils_1.logger.warn(e);
    }
    return false;
}
function replaceRecursive(rv, rule, processor) {
    if (!rule.start || !rule.end) {
        throw new Error('Recursive rule must have start and end.');
    }
    const startRegex = new RegExp(rule.start, 'mi');
    const endRegex = new RegExp(rule.end, 'mi');
    function matchReplacePass(content) {
        const matches = xregexp_1.default.matchRecursive(content, rule.start, rule.end, 'gmi', {
            valueNames: ['between', 'left', 'match', 'right']
        });
        // 如果未命中则直接返回内容
        if (matches.length === 0)
            return content;
        const matchGroup = {
            left: null,
            match: null,
            right: null
        };
        return matches.reduce(function (builder, match) {
            switch (match.name) {
                case 'between':
                    builder += match.value;
                    break;
                case 'left':
                    matchGroup.left = startRegex.exec(match.value);
                    break;
                case 'match':
                    matchGroup.match = match.value;
                    break;
                case 'right':
                    matchGroup.right = endRegex.exec(match.value);
                    builder += processor(matchGroup.left, matchGroup.right, matchGroup.match, matchReplacePass);
                    break;
            }
            return builder;
        }, '');
    }
    return matchReplacePass(rv);
}
function preprocessor(src, context, opts, noRestoreEol, filePath) {
    let rv = normalizeEol(src);
    const processor = (startMatches, endMatches, include, recurse) => {
        if (!startMatches || !endMatches || !include)
            return '';
        const variant = startMatches[1];
        const test = (startMatches[2] || '').trim();
        switch (variant) {
            case 'if': {
                let testResult = testPasses(test, context);
                // 当前传入的 context 没有该 key
                if (testResult instanceof ReferenceError) {
                    utils_1.logger.warn('当前条件编译中找不到变量，将按照条件执行结果为 false 处理\n' +
                        `条件判断语句: ${test}\n` +
                        `报错信息: ${testResult.message}` +
                        (filePath ? `\n文件路径: ${filePath}` : ''));
                }
                if (typeof testResult !== 'boolean')
                    testResult = false;
                return testResult ? recurse(include) : '';
            }
            case 'ifdef':
                return typeof getDeepPropFromObj(context, test) !== 'undefined'
                    ? recurse(include)
                    : '';
            case 'ifndef':
                return typeof getDeepPropFromObj(context, test) === 'undefined'
                    ? recurse(include)
                    : '';
            default:
                throw new Error('Unknown if variant ' + variant + '.');
        }
    };
    if (opts.type.if)
        rv = replaceRecursive(rv, opts.type.if, processor);
    if (!noRestoreEol)
        rv = restoreEol(rv, opts.srcEol);
    return rv;
}
function getEolType(source) {
    let eol;
    let foundEolTypeCnt = 0;
    if (source.indexOf('\r\n') >= 0) {
        eol = '\r\n';
        foundEolTypeCnt++;
    }
    if (/\r[^\n]/.test(source)) {
        eol = '\r';
        foundEolTypeCnt++;
    }
    if (/[^\r]\n/.test(source)) {
        eol = '\n';
        foundEolTypeCnt++;
    }
    if (eol == null || foundEolTypeCnt > 1)
        eol = os_1.EOL;
    return eol;
}
function preprocess(sourceCode, context, ext, filePath) {
    let type;
    if (JsLikeFileExts.includes(ext)) {
        type = 'js';
    }
    else if (XmlLikeFileExts.includes(ext)) {
        type = 'xml';
    }
    if (!type)
        return sourceCode;
    return preprocessor(sourceCode, context, { type: RegexRules[type], srcEol: getEolType(sourceCode) }, undefined, filePath);
}
exports.preprocess = preprocess;
//# sourceMappingURL=codePreprocessor.js.map